@@ -15,27 +15,26 @@
 # You should have received a copy of the GNU General Public License along
 # with sandwine. If not, see <https://www.gnu.org/licenses/>.
 
-import glob
 import logging
 import os
 import random
 import shlex
-import shutil
 import signal
 import subprocess
 import sys
-import time
 from argparse import ArgumentParser, RawTextHelpFormatter
 from contextlib import nullcontext
 from dataclasses import dataclass
 from enum import Enum, auto
+from importlib.metadata import metadata
 from operator import attrgetter, itemgetter
 from textwrap import dedent
 from typing import Optional
 
 import coloredlogs
 
-from sandwine._metadata import DESCRIPTION, VERSION
+from sandwine._pty import pty_spawn_argv
+from sandwine._x11 import X11Display, X11Mode, create_x11_context, detect_and_require_nested_x11
 
 _logger = logging.getLogger(__name__)
 
@@ -45,20 +44,6 @@ class AccessMode(Enum):
     READ_WRITE = 'rw'
 
 
-class X11Mode(Enum):
-    AUTO = 'auto'
-    HOST = 'host'
-    NXAGENT = 'nxagent'
-    XEPHYR = 'xephyr'
-    XNEST = 'xnest'
-    XVFB = 'xvfb'
-    NONE = 'none'
-
-    @staticmethod
-    def values():
-        return [x.value for x in X11Mode.__members__.values()]
-
-
 class MountMode(Enum):
     DEVTMPFS = auto()
     BIND_RO = auto()
@@ -69,6 +54,8 @@ class MountMode(Enum):
 
 
 def parse_command_line(args):
+    distribution = metadata('sandwine')
+
     usage = dedent('''\
         usage: sandwine [OPTIONS] [--] PROGRAM [ARG ..]
            or: sandwine [OPTIONS] --configure
@@ -79,7 +66,7 @@ def parse_command_line(args):
     parser = ArgumentParser(
         prog='sandwine',
         usage=usage,
-        description=DESCRIPTION,
+        description=distribution['Summary'],
         formatter_class=RawTextHelpFormatter,
         epilog=dedent("""\
         Software libre licensed under GPL v3 or later.
@@ -89,7 +76,7 @@ def parse_command_line(args):
     """),
     )
 
-    parser.add_argument('--version', action='version', version=VERSION)
+    parser.add_argument('--version', action='version', version=distribution['Version'])
 
     program = parser.add_argument_group('positional arguments')
     program.add_argument('argv_0', metavar='PROGRAM', nargs='?', help='command to run')
@@ -104,8 +91,8 @@ def parse_command_line(args):
                           dest='x11',
                           action='store_const',
                           const=X11Mode.AUTO,
-                          help='enable nested X11 using X2Go nxagent or Xephry or Xnest'
-                          ' but not Xvfb'
+                          help='enable nested X11 using X2Go nxagent or Xephyr or Xnest'
+                          ' but not Xvfb and not Xpra'
                           ' (default: X11 disabled)')
     x11_args.add_argument('--nxagent',
                           dest='x11',
@@ -116,12 +103,18 @@ def parse_command_line(args):
                           dest='x11',
                           action='store_const',
                           const=X11Mode.XEPHYR,
-                          help='enable nested X11 using Xephry (default: X11 disabled)')
+                          help='enable nested X11 using Xephyr (default: X11 disabled)')
     x11_args.add_argument('--xnest',
                           dest='x11',
                           action='store_const',
                           const=X11Mode.XNEST,
                           help='enable nested X11 using Xnest (default: X11 disabled)')
+    x11_args.add_argument('--xpra',
+                          dest='x11',
+                          action='store_const',
+                          const=X11Mode.XPRA,
+                          help='enable nested X11 using Xpra (EXPERIMENTAL, CAREFUL!)'
+                          ' (default: X11 disabled)')
     x11_args.add_argument('--xvfb',
                           dest='x11',
                           action='store_const',
@@ -146,7 +139,7 @@ def parse_command_line(args):
     mount = parser.add_argument_group('mount arguments')
     mount.add_argument('--dotwine',
                        metavar='PATH:{ro,rw}',
-                       help='use PATH for ~/.wine/ (default: use tmpfs, empty and non-persistant)')
+                       help='use PATH for ~/.wine/ (default: use tmpfs, empty and non-persistent)')
     mount.add_argument('--pass',
                        dest='extra_binds',
                        default=[],
@@ -159,6 +152,13 @@ def parse_command_line(args):
                          action='store_true',
                          help='enforce running winecfg before start of PROGRAM'
                          ' (default: run winecfg as needed)')
+    general.add_argument('--no-pty',
+                         dest='with_pty',
+                         default=True,
+                         action='store_false',
+                         help='refrain from creating a pseudo-terminal'
+                         ', stop protecting against TIOCSTI/TIOCLINUX hijacking (CAREFUL!)'
+                         ' (default: create a pseudo-terminal)')
     general.add_argument('--no-wine',
                          dest='with_wine',
                          default=True,
@@ -175,84 +175,6 @@ def parse_command_line(args):
     return parser.parse_args(args)
 
 
-class X11Context:
-
-    def __init__(self, config):
-        self._config = config
-        self._geometry = '1024x768'
-        self._process = None
-
-    @staticmethod
-    def get_host_display() -> int:
-        try:
-            return int(os.environ['DISPLAY'][1:])
-        except (KeyError, ValueError):
-            return 0
-
-    @staticmethod
-    def find_unused_display() -> int:
-        used_displays = {int(os.path.basename(p)[1:]) for p in glob.glob('/tmp/.X11-unix/X*')}
-        candidate_displays = set(list(range(len(used_displays))) + [len(used_displays)])
-        return sorted(candidate_displays - used_displays)[-1]
-
-    @staticmethod
-    def get_unix_socket_for(display: int) -> str:
-        return f'/tmp/.X11-unix/X{display}'
-
-    @staticmethod
-    def detect_nested():
-        tests = [
-            # NOTE: No Xvfb here because this is meant to be about
-            #       options with user-visible Windows
-            ['nxagent', X11Mode.NXAGENT],
-            ['Xephyr', X11Mode.XEPHYR],
-            ['Xnest', X11Mode.XNEST],
-        ]
-        for command, mode in tests:
-            if shutil.which(command) is not None:
-                _logger.info(f'Using {command} for nested X11.')
-                return mode
-
-        commands = [command for command, _ in tests]
-        _logger.error(f'Neither {" nor ".join(commands)} is available, please install, aborting.')
-        sys.exit(127)
-
-    def __enter__(self):
-        if X11Mode(self._config.x11) != X11Mode.HOST:
-            # NOTE: Extension MIT-SHM is disabled because it kept crashing Xephyr 21.1.7
-            #       and nxagent/X2Go 4.1.0.3 when moving windows around near screen edges.
-            #       PS: Xnest does not support extension MIT-SHM.
-            if X11Mode(self._config.x11) == X11Mode.XEPHYR:
-                argv = ['Xephyr', '-screen', self._geometry, '-extension', 'MIT-SHM']
-            elif X11Mode(self._config.x11) == X11Mode.XNEST:
-                argv = ['Xnest', '-geometry', self._geometry]
-            elif X11Mode(self._config.x11) == X11Mode.NXAGENT:
-                argv = ['nxagent', '-nolisten', 'tcp', '-ac', '-noshmem', '-R']
-            elif X11Mode(self._config.x11) == X11Mode.XVFB:
-                argv = ['Xvfb', '-screen', '0', f'{self._geometry}x24', '-extension', 'MIT-SHM']
-            else:
-                assert False, f'X11 mode {self._config.x11} not supported'
-
-            argv += [f':{self._config.x11_display_number}']
-
-            _logger.info('Starting nested X11...')
-            try:
-                self._process = subprocess.Popen(argv)
-            except FileNotFoundError:
-                _logger.error(f'Command {argv[0]!r} is not available, aborting.')
-                sys.exit(127)
-
-        x11_unix_socket = self.get_unix_socket_for(self._config.x11_display_number)
-        while not os.path.exists(x11_unix_socket):
-            time.sleep(0.5)
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        if self._process is not None:
-            _logger.info('Shutting down nested X11...')
-            self._process.send_signal(signal.SIGINT)
-            self._process.wait()
-
-
 class ArgvBuilder:
 
     def __init__(self):
@@ -273,7 +195,7 @@ class ArgvBuilder:
     def announce_to(self, target):
         for i, group in enumerate(self._groups):
             prefix = '# ' if (i == 0) else ' ' * 4
-            flat_args = ' '.join(shlex.quote(arg) for arg in group)
+            flat_args = shlex.join(group)
             suffix = '' if (i == len(self._groups) - 1) else ' \\'
             print(f'{prefix}{flat_args}{suffix}', file=target)
 
@@ -327,12 +249,12 @@ def create_bwrap_argv(config):
     ]
     env_tasks = {var: None for var in ['HOME', 'TERM', 'USER', 'WINEDEBUG']}
     env_tasks['container'] = 'sandwine'
-    unshare_args = ['--unshare-all']
+    unshare_args = ['--unshare-user', '--unshare-all']
 
     argv = ArgvBuilder()
 
     argv.add('bwrap')
-    argv.add('--new-session')
+    argv.add('--disable-userns')
     argv.add('--die-with-parent')
 
     # Hostname
@@ -356,7 +278,7 @@ def create_bwrap_argv(config):
 
     # X11
     if X11Mode(config.x11) != X11Mode.NONE:
-        x11_unix_socket = X11Context.get_unix_socket_for(config.x11_display_number)
+        x11_unix_socket = X11Display(config.x11_display_number).get_unix_socket()
         mount_tasks += [MountTask(MountMode.BIND_RW, x11_unix_socket)]
         env_tasks['DISPLAY'] = f':{config.x11_display_number}'
 
@@ -477,7 +399,7 @@ def create_bwrap_argv(config):
 
     argv.add('--')
 
-    # Wrap with wineserver (for clean shutdown, it defaults to 3 seconds timout)
+    # Wrap with wineserver (for clean shutdown, it defaults to 3 seconds timeout)
     if config.with_wine:
         argv.add('sh', '-c', 'wineserver -p0 && "$0" "$@" ; ret=$? ; wineserver -k ; exit ${ret}')
 
@@ -501,6 +423,21 @@ def create_bwrap_argv(config):
     return argv
 
 
+def require_recent_bubblewrap():
+    argv = ['bwrap', '--disable-userns', '--help']
+    if subprocess.call(argv, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) != 0:
+        _logger.error('sandwine requires bubblewrap >=0.8.0'
+                      ', aborting.')
+        sys.exit(1)
+
+
+def spawn_argv(argv: list[str], with_pty: bool) -> int:
+    if not with_pty:
+        return subprocess.call(argv)
+
+    return pty_spawn_argv(argv)
+
+
 def main():
     exit_code = 0
     try:
@@ -508,18 +445,23 @@ def main():
 
         coloredlogs.install(level=logging.DEBUG)
 
+        require_recent_bubblewrap()
+
         if X11Mode(config.x11) != X11Mode.NONE:
             if X11Mode(config.x11) == X11Mode.AUTO:
-                config.x11 = X11Context.detect_nested()
+                config.x11 = detect_and_require_nested_x11()
 
             if X11Mode(config.x11) == X11Mode.HOST:
-                config.x11_display_number = X11Context.get_host_display()
+                config.x11_display_number = X11Display.find_used()
             else:
-                config.x11_display_number = X11Context.find_unused_display()
+                minimum = 0
+                if X11Mode(config.x11) == X11Mode.XPRA:
+                    minimum = 10  # Avoids warning from Xpra for displays <=9
+                config.x11_display_number = X11Display.find_unused(minimum)
 
             _logger.info('Using display ":%s"...', config.x11_display_number)
 
-            x11context = X11Context(config)
+            x11context = create_x11_context(config.x11, config.x11_display_number, 1024, 768)
         else:
             x11context = nullcontext()
 
@@ -530,7 +472,7 @@ def main():
 
         with x11context:
             try:
-                exit_code = subprocess.call(argv)
+                exit_code = spawn_argv(argv, with_pty=config.with_pty)
             except FileNotFoundError:
                 _logger.error(f'Command {argv[0]!r} is not available, aborting.')
                 exit_code = 127
